#!/usr/bin/env python

#    Copyright (C) 2003 Paul Harrison
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""

  Little Snob

    Little Snob is a Python data clustering library, based on the original
    Fortran and C Snob software by Chris Wallace.

    This software has not been fully tested. If in doubt, check results
    against the original Snobs.


    As input, Little Snob requires a list of samples. Each sample may contain
    several properties, which may either be continuous valued or discrete.
    
    As output, Little Snob will produce a list of classes that may
    (hopefully) be used to encode this list in the fewest possible number
    of bits. Each class will specify the distribution of sample properties
    within that class. Continuous valued properties are modelled using
    a normal distribution, while discrete valued properties are modelled
    by specifying a probability for each possible value.

    The number of classes to be output does *not* need to be specified. Little
    Snob will choose the correct number of classes.

    Little Snob can also output for each sample the class that it is most likely
    to be a member of. It is also possible to obtain the probability that a
    particular sample is a member of a particular class, if this file is used as
    a library in another program.


    Little Snob has most of the features of "Vanilla Snob". It may be
    slightly slower to find the best set of classes however.

    Little Snob is quite a lot shorter than Vanilla Snob, mostly on
    account of being written in Python. Those of you familiar with the
    theory of MML may now chuckle quietly.

    A number of non-free variants of Snob exist with other features,
    notably factor analysis. If someone wants to add factor analysis to
    this variant, that would be really nifty.


  History

    0.1 -- Initial release

    0.2 -- Added "-a" option
           Removed "select" (should now work on win32)
           Read files with universal newline ("U") mode

    0.3 -- Support for a "significance" bit (as per Vanilla Snob)
               (see Possibly_uncompressable_column)
           Correct partial assignement (?)
           More comments
           Some minor performance tweaks and bug fixes
    
"""

__version__ = "0.3"

# TODO: Gaussian_column density should adjust for quantum
#       (probably not a big deal)

from __future__ import generators

import random, string, sys, signal, getopt, time

import Numeric as N

# Basic distributions ==========

class Column:
    """ This is an abstract base class.

        Imagine the data we are compressing as a table, with each row 
        representing a thing, and each column a property of those things.
    
        Subclasses of this class provide a way to compress subsets of
        one column of the data.
    """

    def __init__(self, name):
        self._name = name

    def get_name(self):
        return self._name


class Multinomial_column( Column ):
    """ A class for compressing discrete (qualitative) data.

        Data items that are "None" are treated as missing data, 
        and require zero bits to store.

        Example: Multinomial_column('species', 
                    ['duck', 'rabbit', 'duck', 'duck', None, 'rabbit'])
    """

    def __init__(self, name, data):
        Column.__init__(self,name)

        # Work out how many different values there are
        dict = { }
        for datum in data:
            if datum is not None:
                dict[datum] = 1

        self._n_values = len(dict)
        
        # Map each value to an integer
        self._values = dict.keys()
        self._values.sort()

        mapping = { }
        for i in xrange(self._n_values):
            mapping[self._values[i]] = i
        mapping[None] = self._n_values

        # Check for any missing data items
        self._valid = N.not_equal(data, None)
        self._all_valid = N.alltrue(self._valid)
        self._data = [ mapping[item] for item in data ]

        # Precompute some stuff for the compress function

        self._takers = [ [ ] for i in xrange(self._n_values) ]
        for i in xrange(len(self._data)):
            if self._valid[i]:
                self._takers[self._data[i]].append(i)
        self._takers = [ N.array(item) for item in self._takers ]

        self._log_h = N.sum(N.log(N.arange(1,self._n_values)))

    def get_null_model(self):
        """ The model in the absence of any data. """

        return N.ones(self._n_values, N.Float) / self._n_values

    def describe_model(self, model):
        """ Represent a model returned by "compress" as a string. """

        list = [ ]
        for i in xrange(self._n_values):
            list.append("%4.1f%% "%(model[i]*100.)+str(self._values[i]))

        return ("%25s: "%self._name) + string.join(list,', ')

    def snob_vset_line(self):
        return self._name + " 2 " + str(self._n_values)

    def snob_samp_column(self):
        return [ [str(item+1),'='][item == self._n_values] for item in self._data ]

    def snob_samp_param(self):
        return ''
 
    def log_densities(self, model):
        """ Log-likelyhood of each data item, given the model """
    
        log_model = N.log(list(model)+[1.])
        result = N.take(log_model, self._data)

        return result

    def compress(self, weighting):
        """ Determine the optimal model and the number of bits required to
            store the model and data, for a sub-set of the column defined by
            "weighting".

            Note: it's a bit of a hack to use values other than 0 and 1 in
                  weighting. But some very smart people say it's ok, so don't
                  sweat it. """
    
        # Distributes weighting by the value of the data,
        # then sums the weights for each value
        counts = [ N.sum(N.take(weighting, self._takers[i])) 
                        for i in xrange(self._n_values) ]

        counts = N.array(counts)

        total = N.sum(counts)

        model = (counts+.5) / (total+.5*self._n_values)
        
        log_model = N.log(model)
        log_f = N.sum(log_model * counts)
        log_F = (self._n_values-1)*N.log(total) - N.sum(log_model)
        
        length = -self._log_h -log_f + .5*log_F + \
                 (self._n_values-1)*(0.5-0.5*N.log(12))

        return length, model


class Gaussian_column( Column ):
    """ A class for compressing continuous valued (quantitative) data in terms 
        of a normal distribution.

        The "quantum" parameter specifies the accuracy of measurement of 
        the data. If in doubt, a very small value will probably do no harm.

        Data items that are "None" are treated as missing data, 
        and require zero bits to store.

        Example: Gaussian_column('height', [70, 80.5, None, 108], 0.1)
    """
    
    def __init__(self, name, data, quantum):
        Column.__init__(self, name)

        self._valid = N.not_equal(data, None)
        self._all_valid = N.alltrue(self._valid)
        self._data = N.where(self._valid, data, 0.).astype(N.Float)

        self._quantum = quantum   # Measurement accuracy x +/- .5*quantum

        self._good_data = N.compress(self._valid, self._data)

        mean, variance, total = self._mean_variance_total(N.ones(len(data), N.Float))
        self._mean = mean
        self._std_dev = N.sqrt(variance)
        self._q_std_dev = self._std_dev / self._quantum

    def log_densities(self, model):
        mean, std_dev = model

        # TODO: correct for quantum

        result = self._data - mean
        result *= result
        result *= -.5/std_dev/std_dev

        q_std_dev = std_dev / self._quantum

        result += N.log(1./(N.sqrt(2.*N.pi)*q_std_dev))

        if not self._all_valid:
            result = N.where(self._valid, result, 0.)

        return result

    def describe_model(self, model):
        return "%25s: mean: %.2f, standard deviation: %.2f" % (self._name, model[0], model[1])

    def get_null_model(self):
        return (self._mean, self._std_dev)

    def snob_vset_line(self):
        return self._name + " 1"

    def snob_samp_column(self):
        return [ ['===',str(self._data[i])][self._valid[i]] \
                     for i in xrange(len(self._data)) ]

    def snob_samp_param(self):
        return str(self._quantum)
 
    def _mean_variance_total(self, weighting):
        if not self._all_valid:
            weighting = N.compress(self._valid, weighting)

        total = N.sum(weighting)

        mean = N.sum(self._good_data*weighting) / total

        # Average over the quantum
        lower = self._good_data -(mean+.5*self._quantum)
        upper = self._good_data -(mean-.5*self._quantum)

        variance = N.sum( (upper*upper*upper-lower*lower*lower)*weighting ) \
                   / (3.*self._quantum*(total-1.))

        return mean, variance, total

    def compress(self, weighting):
        mean, variance, total = self._mean_variance_total(weighting)

        # In terms of quantum units
        q_variance = variance / self._quantum

        # TODO: check this!
        length = N.log(self._q_std_dev) + N.log(total) - .5*N.log(2) \
                 - N.log(q_variance) + .5 \
                 + .5*total*(N.log(2*N.pi)+N.log(q_variance)+.5)

        return length, (mean, N.sqrt(variance))



class Possibly_uncompressable_column( Column ):
    """ A column that might be compressed by a certain model,
        or might be best just stored with no model.
        
        Instances of this class wrap around another Column type object 
        that does the actual compression.
        
        Example: 
            Possibly_uncompressable_column(
                Multinomial_column('species', 
                    ['duck', 'rabbit', 'duck', 'duck', 'fox', 'rabbit']))
    """

    def __init__(self, column):
        Column.__init__(self, column.get_name())

        self._column = column

        self._null_model = column.get_null_model()
        self._null_log_densities = column.log_densities(self._null_model)

    def describe_model(self, model):
        if model[0]:
            return self._column.describe_model(model[1])
        else:
            return "%25s: not significant" % self._name

    def snob_vset_line(self):
        return self._column.snob_vset_line()

    def snob_samp_column(self):
        return self._column.snob_samp_column()

    def snob_samp_param(self):
        return self._column.snob_samp_param()

    def log_densities(self, model):
        if model[0]:
            return self._column.log_densities(model[1])
        else:
            return self._null_log_densities
 
    def compress(self, weighting):
        """ We have two hypotheses:
              H0 - the data can not be compressed
              H1 - the data can be compressed

              Choose the most likely (to transmit this choice takes one bit). """

        H1_length, H1_model = self._column.compress(weighting)

        H0_length = -N.sum(self._null_log_densities * weighting)

        if H0_length < H1_length:
            return H0_length+1., (0, self._null_model)
        else:
            return H1_length+1., (1, H1_model)


# Mixture model ==========

class Mixture_model:
    """ Model object returned by "refine_mixture_model".

        model.class_p is the prior probability of a random sample belonging to
                       each class.

        model.class_models is a list of the class models. Each class model
                       comprises a list of column models. """

    def __init__(self, class_p, class_models, columns):
        self.class_p = class_p
        self.class_models = class_models
        self.columns = columns

    def __repr__(self):
        str =  ''

        for i in xrange(len(self.class_models)):
            str += "%4.1f%% class %d\n" % (self.class_p[i]*100., i+1)
            for j in xrange(len(self.columns)):
                str += self.columns[j].describe_model(self.class_models[i][j])+'\n'

            str += '\n'

        return str


def _make_model(columns, assignments):
    """ Returns a mixture model, and its length. """

    n_classes = len(assignments)
    
    # TODO: don't compute this twice (also in prune)
    totals = [ N.sum(item) for item in assignments ]

    class_compressor = Multinomial_column('', range(n_classes))

    # The first part of the message is the assignment of samples to
    # classes:

    length, class_p = class_compressor.compress(totals)

    # Partial assignment adjustment, assignment not fully specified in encoding,
    # so delete the "nuisance bits":

    length -= -N.sum(N.log(class_p) * totals)
    length += -N.sum(N.log(N.sum(N.swapaxes(assignments,0,1)*class_p, 1)))
    
    # ^^ notionally: if a sample might be a member of several classes
    #       we assign it to a class randomly. We lump these different possible
    #       messages together as equivalent, the expected message length is 
    #       shorter if it doesn't distinguish between them.
    #
    #    This should let snob distinguish classes that are close together,
    #    or even differ only in variance.

    # Don't care about model order, thus subtract out the bits required
    # to specify it:
    
    length -= N.sum(N.log(N.arange(1,n_classes+1)))


    # The rest of the message is the encoding of each class
    # (both the class model and the samples in that class)
    
    class_models = [ ]
    for i in xrange(n_classes):
        class_model = [ ]
        for j in xrange(len(columns)):
            column_length, column_model = \
                columns[j].compress(assignments[i])
                
            class_model.append(column_model)
            length += column_length
        
        class_models.append(class_model)


    return Mixture_model(class_p, class_models, columns), length


def _make_assignments(model, data_size):
    """ Assigns samples to classes in a mixture model. """

    # Work in logs to avoid underflows
    # (eg when there are a large number of columns and the densitiy becomes
    #     tiny)
    log_densities = N.repeat(N.log(model.class_p[:,N.NewAxis]), data_size, 1)

    for i in xrange(len(model.class_models)):
        log_densities[i] += N.sum([
          model.columns[j].log_densities(model.class_models[i][j])
            for j in xrange(len(model.columns))
        ])

    # Avoid underflow
    log_densities -= N.maximum.reduce(log_densities)
    log_densities = N.maximum(log_densities, -50.)

    densities = N.exp(log_densities)
    densities /= N.sum(densities)

    return densities


def _split_a_class(assignments, model, columns):
    """ Split a randomly chosen class in two, randomly. """

    class_index = random.randrange(len(assignments))

    new = list(assignments)
    
    a = new[class_index].copy()
    new[class_index] = a
    
    b = a.copy()
    new.append(b)

    amount = random.random()
    for i in xrange(len(a)):
        if random.random() < amount:
            a[i] = 0.
        else:
            b[i] = 0.

    return N.array(new)


def _join_two_classes(assignments, which=None):
    """ Merge a randomly chosen class into the class it is most similar too. """

    # TODO: a better measure of similarity

    if which:
        index_1 = which
    else:
        index_1 = random.randrange(len(assignments))

    index_2 = 0
    best_distance2 = 1e30
    for i in xrange(len(assignments)):
        if i != index_1:
           difference = assignments[index_1] - assignments[i]
           distance2 = N.dot(difference,difference)
           if distance2 < best_distance2:
               best_distance2 = distance2
               index_2 = i

    new = list(assignments)
    new[index_1] = new[index_1] + new[index_2]
    del new[index_2] 

    return N.array(new)


def _prune_classes(assignments):
    """ Remove classes with size less than 2. """

    while 1:
        for i in xrange(len(assignments)):
            if N.sum(assignments[i]) < 2.:
                assignments = _join_two_classes(assignments, i)
                break
        else:
            return assignments


def default_status_func(iter, n_iter, current, best):
    """ Default status function used by "refine_mixture_model",
        prints out current status to stderr. """

    sys.stderr.write(
     "\r%5d  Current: %.1f bits, %2d classes.  Best: %.1f bits, %2d classes.  " 
     % ( iter+1, current[0]/N.log(2), len(current[1].class_models), 
         best[0]/N.log(2), len(best[1].class_models)))
    sys.stderr.flush()


def refine_mixture_model(columns, data_size, iters=1000, \
                         assignments=None, status_func=default_status_func):

    """ Create or refine a mixture model.

        This corresponds to "doall" in the original Snob.

        Returns a tuple of the length, model and assignments for the best model
        found.

        Parameters:

          columns is a list of data columns (which must be subclasses of the 
              Column class).

          data_size is the number of samples.

          iters is the number of iterations to perform (use sys.maxint to
              iterate indefinitely).

          assignments, if specified, is the value returned from a previous
              invocation of refine_mixture_model.

          status_func, if not None, is called at the end of each iteration.
              If this function returns true, iteration will stop early.
    """
    
    if assignments is None:
        assignments = N.repeat([[1.]], data_size, 1)

    assignments = _prune_classes(assignments)
    model, length = _make_model(columns, assignments)

    best_length = length
    best_model = model
    best_assignments = assignments

    # Strategy is this:
    #   Keep a copy of the best model found so far
    #   - don't look at recently mutated models,
    #     let them settle down a bit first
    #
    #   Sometimes revert to the best model unless 
    #   - the current model is changing in length
    #   - the current model has just been mutated
    #   but sometimes revert just for the hell of it
    #
    #   Sometimes mutate the model unless 
    #   - the current model is changing in length
    #   - the current model has just been mutated
    #   but sometimes just mutate for the hell of it
    #
    #   If the current model has just become the best
    #   model so far, almost always mutate it immediately
    #   (we have a backup now, so can fine tune it later
    #    if further mutation fails)
    #
    #   The meat of the loop is to refine the current model,
    #   by tuning the model parameters, then tuning the assignments
    #   of samples to classes, repeatedly.

    changing = 1
    settling_countdown = 0

    for i in xrange(iters):
        if not settling_countdown and \
           random.random() < [0.9,0.01][changing]:
            length = best_length
            model = best_model
            assignments = best_assignments
            settling_countdown = 0
            changing = 0

        assignments = _make_assignments(model, data_size)

        if not settling_countdown and \
           random.random() < [0.5,0.01][changing]:
            settling_countdown = 5
            if random.randrange(4) == 0 and len(assignments) > 1:
                assignments = _join_two_classes(assignments)
            else:
                assignments = _split_a_class(assignments, model, columns)
        
        assignments = _prune_classes(assignments)

        old_length = length
        model, length = _make_model(columns, assignments)

        changing = (abs(old_length-length) > 1.)
        settling_countdown = max(0,settling_countdown-1)

        if not settling_countdown and \
           length < best_length:
            best_length = length
            best_model = model
            best_assignments = assignments
            changing = 0

        if status_func:
            halt = status_func(i, iters, (length,model,assignments), (best_length,best_model,best_assignments) )
            if halt: break

    return best_length, best_model, best_assignments


# SNOB format file loading ===========

def _file_iter(filename):
    file = open(filename, "rU")

    for line in file.xreadlines():
        for item in line.split():
            yield item

    file.close()

def _load_vset(filename):
    lines = open(filename,"rU").readlines()
    lines = [ line for line in lines if line.strip() ]
    n_classes = int(lines[1].strip())
    list = [ ]
    for i in xrange(n_classes):
        split = lines[i+2].split()
        if int(split[1]) == 1:
            type = Gaussian_column
        else:
            type = Multinomial_column
        list.append((split[0], type))

    return list

def _load_samp(filename, vset):
    iter = _file_iter(filename)

    # Sample set name
    iter.next()

    # Vset name
    iter.next()

    quanta = { }
    for i in range(len(vset)):
        if vset[i][1] is Gaussian_column:
            quanta[i] = float( iter.next() )
    
    n_samples = int( iter.next() )

    indicies = [ ]
    columns = [ [] for item in vset ]

    for i in xrange(n_samples):
        indicies.append( iter.next() )
        for i in range(len(vset)):
            item = iter.next()

            if item == '='*len(item):
                columns[i].append(None)
            elif vset[i][1] is Gaussian_column:
                columns[i].append(float(item)) 
            else:
                columns[i].append(item)

    try:
        iter.next()
        raise "Sample file incorrectly formatted"
    except StopIteration:
        pass

    column_compressors = [ ]
    for i in range(len(vset)):
        if vset[i][1] is Gaussian_column:
            column_compressors.append(Gaussian_column(vset[i][0], columns[i], quanta[i]))
        else:
            column_compressors.append(vset[i][1](vset[i][0], columns[i]))

    # Vanilla Snob has a "is it significant" bit
    column_compressors = [ Possibly_uncompressable_column(item) 
                           for item in column_compressors ]

    return indicies, column_compressors

def write_snob_files(prefix, indicies, columns):
    """ Write data suitable for reading by Vanilla Snob or snob.py invoked 
        from the command line.

        This might be useful if you want to compare the results of Vanilla and
        Little Snob, and you are using Little Snob as a library rather than a
        program. """

    file = open(prefix + '.vset', 'wt')

    file.write('%s\n%d\n' % (prefix, len(columns)))

    for column in columns:
        file.write(column.snob_vset_line()+'\n')

    file.close()

    file = open(prefix + '.samp', 'wt')

    file.write('%s\n%s\n' % (prefix,prefix))

    for column in columns:
        file.write(column.snob_samp_param()+'\n')
    
    file.write('%d\n' % len(indicies))

    samp_columns = [ column.snob_samp_column() for column in columns ] 
    indicies = [ str(item) for item in indicies ]
    lines = zip(indicies, *samp_columns)

    for line in lines:
       file.write(string.join(line,'\t')+'\n')

    file.close()


# Simple interface ==========

# If you want something fancy and graphical, write it :-)

def refine_until_stopped(columns, n_data):
    """ Iterate refine_mixture_model until the user presses ctrl-c """

    print >> sys.stderr, "\nClassifying", n_data, "samples."
    print >> sys.stderr, "Press ctrl-c to finish.\n"

    interrupted = [0]
    last_call = [0]

    def int_handler(signal, frame):
        interrupted[0] = 1

    def status_func(*args):
        if interrupted[0]:
            return 1

        now = time.time()
        if abs(now-last_call[0]) < .1:
            return
        last_call[0] = now

        default_status_func(*args)

    signal.signal(signal.SIGINT, int_handler)

    result = refine_mixture_model(columns, n_data, sys.maxint, status_func=status_func)
    
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    print >> sys.stderr, "\n"

    return result


def display_result(indicies, length, model, assignments, 
                   show_assignments=0, file=sys.stdout):

    print >> file, "The model and data may together be stored in %.1f bits," % (length/N.log(2))
    print >> file, "if the model comprises the following", len(model.class_models), "classes:"
    print >> file
    print >> file, model,

    if show_assignments:
        assignments = N.argmax(assignments,0)

        for i in xrange(len(model.class_models)):
            print >> file
            print >> file, "Class %d:" % (i+1)
            print >> file
            position = 0
            for j in xrange(n_data):
                 if assignments[j] == i:
                     desc = str(indicies[j])
                     if position + len(desc)+2 >= 70:
                         print >> file
                         position = 0
                     file.write('  '+desc)

                     position += len(desc)+2
            print >> file


          
if __name__ == '__main__':
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'ah', ['--help'])
    except getopt.GetoptError:
        opts = [('-h','')]

    show_assignments = 0

    if len(args) != 2:
        opts = [('-h','')]

    for option, parameter in opts:
        if option in ('-h', '--help'):
            print >> sys.stderr, \
"""
Usage:

  python snob.py [options] <vset file> <samp file>

  Options:

      -a   print out assignments of samples to classes

"""
            sys.exit(1)

        elif option == '-a':
            show_assignments = 1

    vset = _load_vset(args[0])
    indicies, columns = _load_samp(args[1], vset)
    n_data = len(indicies)

    best_length, best_model, best_assignments = refine_until_stopped(columns, n_data)

    display_result(indicies, best_length, best_model, best_assignments, show_assignments)

